from __future__ import annotations

import wrapt
from matplotlib.axes import Axes
from matplotlib.lines import Line2D

from maidr.core.enum import PlotType
from maidr.patch.common import common


def line(wrapped, instance, args, kwargs) -> Axes | list[Line2D]:
    """
    Wrapper function for line plotting functions in matplotlib and seaborn.

    Parameters
    ----------
    wrapped : callable
        The wrapped function (plot or lineplot)
    instance : object
        The object to which the wrapped function belongs
    args : tuple
        Positional arguments passed to the wrapped function
    kwargs : dict
        Keyword arguments passed to the wrapped function

    Returns
    -------
    Axes | list[Line2D]
        The result of the wrapped function after processing
    """
    return common(PlotType.LINE, wrapped, instance, args, kwargs)


# Patch matplotlib function.
wrapt.wrap_function_wrapper(Axes, "plot", line)

# Patch seaborn function.
wrapt.wrap_function_wrapper("seaborn", "lineplot", line)



from __future__ import annotations

import wrapt

from matplotlib.axes import Axes
from matplotlib.image import AxesImage

from maidr.core.enum import PlotType
from maidr.core.figure_manager import FigureManager


def heat(wrapped, _, args, kwargs) -> Axes | AxesImage:
    # Check for additional params used by MAIDR heatmap.
    optional_params = {}
    if "fill_label" in kwargs:
        # Remove `fill_label` because it is introduced by us.
        optional_params["fill_label"] = kwargs.pop("fill_label")
    if "fmt" in kwargs:
        optional_params["fmt"] = kwargs["fmt"]

    # Patch `ax.imshow()` and `seaborn.heatmap`.
    plot = wrapped(*args, **kwargs)

    # Extract the heatmap data points for MAIDR from the plots.
    ax = FigureManager.get_axes(plot)
    FigureManager.create_maidr(ax, PlotType.HEAT, **optional_params)

    # Return to the caller.
    return plot


# Patch matplotlib function.
wrapt.wrap_function_wrapper(Axes, "imshow", heat)

# Patch seaborn function.
wrapt.wrap_function_wrapper("seaborn", "heatmap", heat)



from __future__ import annotations

import io
import json
import os
import tempfile
import uuid
import webbrowser
from typing import Literal

from htmltools import HTML, HTMLDocument, Tag, tags
from lxml import etree
from matplotlib.figure import Figure

from maidr.core.context_manager import HighlightContextManager
from maidr.core.enum.maidr_key import MaidrKey
from maidr.core.enum.plot_type import PlotType
from maidr.core.plot import MaidrPlot
from maidr.util.environment import Environment


class Maidr:
    """
    A class to handle the rendering and interaction of matplotlib figures with additional metadata.

    Attributes
    ----------
    _fig : Figure
        The matplotlib figure associated with this instance.
    _plots : list[MaidrPlot]
        A list of MaidrPlot objects which hold additional plot-specific configurations.

    Methods
    -------
    render(lib_prefix=None, include_version=True)
        Creates and returns a rendered HTML representation of the figure.
    save_html(file, lib_dir=None, include_version=True)
        Saves the rendered HTML representation to a file.
    show(renderer='auto')
        Displays the rendered HTML content in the specified rendering context.
    """

    def __init__(self, fig: Figure, plot_type: PlotType = PlotType.LINE) -> None:
        """Create a new Maidr for the given ``matplotlib.figure.Figure``."""
        self._fig = fig
        self._plots = []
        self.maidr_id = None
        self.selector_id = Maidr._unique_id()
        self.plot_type = plot_type

    @property
    def fig(self) -> Figure:
        """Return the ``matplotlib.figure.Figure`` associated with this object."""
        return self._fig

    @property
    def plots(self) -> list[MaidrPlot]:
        """Return the list of plots extracted from the ``fig``."""
        return self._plots

    def render(self) -> Tag:
        """Return the maidr plot inside an iframe."""
        return self._create_html_tag()

    def save_html(
        self, file: str, *, lib_dir: str | None = "lib", include_version: bool = True
    ) -> str:
        """
        Save the HTML representation of the figure with MAIDR to a file.

        Parameters
        ----------
        file : str
            The file to save to.
        lib_dir : str, default="lib"
            The directory to save the dependencies to
            (relative to the file's directory).
        include_version : bool, default=True
            Whether to include the version number in the dependency folder name.
        """
        html = self._create_html_doc()
        return html.save_html(file, libdir=lib_dir, include_version=include_version)

    def show(
        self,
        renderer: Literal["auto", "ipython", "browser"] = "auto",
    ) -> object:
        """
        Preview the HTML content using the specified renderer.

        Parameters
        ----------
        renderer : Literal["auto", "ipython", "browser"], default="auto"
            The renderer to use for the HTML preview.
        """
        html = self._create_html_tag()
        _renderer = Environment.get_renderer()
        if _renderer == "browser" or (
            Environment.is_interactive_shell() and not Environment.is_notebook()
        ):
            return self._open_plot_in_browser()
        return html.show(renderer)

    def clear(self):
        self._plots = []

    def destroy(self) -> None:
        del self._plots
        del self._fig

    def _open_plot_in_browser(self) -> None:
        """
        Open the rendered HTML content using a temporary file
        """
        system_temp_dir = tempfile.gettempdir()
        static_temp_dir = os.path.join(system_temp_dir, "maidr")
        os.makedirs(static_temp_dir, exist_ok=True)

        temp_file_path = os.path.join(static_temp_dir, "maidr_plot.html")
        html_file_path = self.save_html(temp_file_path)
        webbrowser.open(f"file://{html_file_path}")

    def _create_html_tag(self) -> Tag:
        """Create the MAIDR HTML using HTML tags."""
        tagged_elements = [element for plot in self._plots for element in plot.elements]
        with HighlightContextManager.set_maidr_elements(tagged_elements):
            svg = self._get_svg()
        maidr = f"\nlet maidr = {json.dumps(self._flatten_maidr(), indent=2)}\n"

        # In SVG we will replace maidr=id with the unique id.
        svg = svg.replace('maidr="true"', f'maidr="{self.selector_id}"')

        # Inject plot's svg and MAIDR structure into html tag.
        return Maidr._inject_plot(svg, maidr, self.maidr_id)

    def _create_html_doc(self) -> HTMLDocument:
        """Create an HTML document from Tag objects."""
        return HTMLDocument(self._create_html_tag(), lang="en")

    def _flatten_maidr(self) -> dict | list[dict]:
        """Return a single plot schema or a list of schemas from the Maidr instance."""
        if self.plot_type == PlotType.LINE:
            self._plots = [self._plots[0]]
        maidr = [plot.schema for plot in self._plots]

        # Replace the selector having maidr='true' with maidr={self.maidr_id}
        for plot in maidr:
            if MaidrKey.SELECTOR in plot:
                plot[MaidrKey.SELECTOR] = plot[MaidrKey.SELECTOR].replace(
                    "maidr='true'", f"maidr='{self.selector_id}'"
                )

        return maidr if len(maidr) != 1 else maidr[0]

    def _get_svg(self) -> HTML:
        """Extract the chart SVG from ``matplotlib.figure.Figure``."""
        svg_buffer = io.StringIO()
        self._fig.savefig(svg_buffer, format="svg")
        str_svg = svg_buffer.getvalue()

        etree.register_namespace("svg", "http://www.w3.org/2000/svg")
        tree_svg = etree.fromstring(str_svg.encode(), parser=None)
        root_svg = None
        # Find the `svg` tag and set unique id if not present else use it.
        for element in tree_svg.iter(tag="{http://www.w3.org/2000/svg}svg"):
            _id = Maidr._unique_id()
            self._set_maidr_id(_id)
            if "id" not in element.attrib:
                element.attrib["id"] = _id
            if "maidr-data" not in element.attrib:
                element.attrib["maidr-data"] = json.dumps(
                    self._flatten_maidr(), indent=2
                )
            root_svg = element
            break

        svg_buffer = io.StringIO()  # Reset the buffer
        svg_buffer.write(
            etree.tostring(
                root_svg, pretty_print=True, encoding="unicode"  # type: ignore
            )
        )

        return HTML(svg_buffer.getvalue())

    def _set_maidr_id(self, maidr_id: str) -> None:
        """Set a unique identifier to each ``MaidrPlot``."""
        self.maidr_id = maidr_id
        for maidr in self._plots:
            maidr.set_id(maidr_id)

    @staticmethod
    def _unique_id() -> str:
        """Generate a unique identifier string using UUID4."""
        return str(uuid.uuid4())

    @staticmethod
    def _inject_plot(plot: HTML, maidr: str, maidr_id) -> Tag:
        """Embed the plot and associated MAIDR scripts into the HTML structure."""

        engine = Environment.get_engine()

        MAIDR_TS_CDN_URL = "https://cdn.jsdelivr.net/npm/maidr-ts/dist/maidr.js"

        maidr_js_script = f"""
            if (!document.querySelector('script[src="https://cdn.jsdelivr.net/npm/maidr/dist/maidr.min.js"]')) {{
                var script = document.createElement('script');
                script.type = 'text/javascript';
                script.src = 'https://cdn.jsdelivr.net/npm/maidr/dist/maidr.min.js';
                script.addEventListener('load', function() {{
                    window.init("{maidr_id}");
                }});
                document.head.appendChild(script);
            }} else {{
                window.init("{maidr_id}");
            }}
        """

        maidr_ts_script = f"""
            if (!document.querySelector('script[src="{MAIDR_TS_CDN_URL}"]'))
            {{
                var script = document.createElement('script');
                script.type = 'module';
                script.src = '{MAIDR_TS_CDN_URL}';
                script.addEventListener('load', function() {{
                    window.main();
                }});
                document.head.appendChild(script);
            }} else {{
                document.addEventListener('DOMContentLoaded', function (e) {{
                    window.main();
                }});
            }}
        """

        script = maidr_js_script if engine == "js" else maidr_ts_script

        base_html = tags.div(
            tags.link(
                rel="stylesheet",
                href="https://cdn.jsdelivr.net/npm/maidr/dist/maidr_style.min.css",
            ),
            tags.script(script, type="text/javascript"),
            tags.div(plot),
        )

        is_quarto = os.getenv("IS_QUARTO") == "True"

        # Render the plot inside an iframe if in a Jupyter notebook, Google Colab
        # or VSCode notebook. No need for iframe if this is a Quarto document.
        # For TypeScript we will use iframe by default for now
        if (Environment.is_notebook() and not is_quarto) or (
            engine == "ts" and Environment.is_notebook()
        ):
            unique_id = "iframe_" + Maidr._unique_id()

            def generate_iframe_script(unique_id: str) -> str:
                resizing_script = f"""
                    function resizeIframe() {{
                        let iframe = document.getElementById('{unique_id}');
                        if (
                            iframe && iframe.contentWindow &&
                            iframe.contentWindow.document
                        ) {{
                            let iframeDocument = iframe.contentWindow.document;
                            let brailleContainer =
                                iframeDocument.getElementById('braille-input');
                            iframe.style.height = 'auto';
                            let height = iframeDocument.body.scrollHeight;
                            if (brailleContainer &&
                                brailleContainer === iframeDocument.activeElement
                            ) {{
                                height += 100;
                            }}else{{
                                height += 50
                            }}
                            iframe.style.height = (height) + 'px';
                            iframe.style.width = iframeDocument.body.scrollWidth + 'px';
                        }}
                    }}
                    let iframe = document.getElementById('{unique_id}');
                    resizeIframe();
                    iframe.onload = function() {{
                        resizeIframe();
                        iframe.contentWindow.addEventListener('resize', resizeIframe);
                    }};
                    iframe.contentWindow.document.addEventListener('focusin', () => {{
                        resizeIframe();
                    }});
                    iframe.contentWindow.document.addEventListener('focusout', () => {{
                        resizeIframe();
                    }});
                """
                return resizing_script

            resizing_script = generate_iframe_script(unique_id)

            base_html = tags.iframe(
                id=unique_id,
                srcdoc=str(base_html.get_html_string()),
                width="100%",
                height="100%",
                scrolling="no",
                style="background-color: #fff; position: relative; border: none",
                frameBorder=0,
                onload=resizing_script,
            )

        return base_html



from __future__ import annotations

from typing import Any

from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable
from matplotlib.lines import Line2D

from maidr.core.enum import MaidrKey


class ContainerExtractorMixin:
    @staticmethod
    def extract_container(
        ax: Axes,
        container_type: type,
        include_all: bool = False,
    ) -> Any:
        """Retrieve containers of a specified type from an Axes object."""
        if ax is None or ax.containers is None:
            return None

        # If include_all is True, return a list of all containers of the specified type.
        if include_all:
            return [
                container
                for container in ax.containers
                if isinstance(container, container_type)
            ]

        # Otherwise, return the first container of the specified type.
        return next(
            container
            for container in ax.containers
            if isinstance(container, container_type)
        )


class LevelExtractorMixin:
    @staticmethod
    def extract_level(ax: Axes, key: MaidrKey = MaidrKey.X) -> list[str] | None:
        """Retrieve label texts from Axes based on the specified Maidr key."""
        if ax is None:
            return None

        level = None
        if MaidrKey.X == key:
            level = [label.get_text() for label in ax.get_xticklabels()]
        elif MaidrKey.Y == key:
            level = [label.get_text() for label in ax.get_yticklabels()]
        elif MaidrKey.FILL == key:
            level = [container.get_label() for container in ax.containers]

        return level


class LineExtractorMixin:
    @staticmethod
    def extract_line(ax: Axes) -> Line2D | None:
        """Retrieve the last line object from Axes, if available."""
        if ax is None or ax.get_lines() is None:
            return None

        # Since the upstream MaidrJS library currently supports only the last plot line,
        # `maidr` package supports the same.
        return ax.get_lines()[-1]

    @staticmethod
    def extract_lines(ax: Axes) -> list[Line2D] | None:
        """Retrieve all line objects from Axes, if available."""
        if ax is None or ax.get_lines() is None:
            return None

        # Since the upstream MaidrJS library currently supports only the last plot line,
        # `maidr` package supports the same.
        return ax.get_lines()


class CollectionExtractorMixin:
    @staticmethod
    def extract_collection(ax: Axes, collection_type: type) -> Any:
        """Retrieve the first collection of a specified type from an Axes object."""
        if ax is None or ax.collections is None:
            return None

        # We assume only one collection of each type is present to avoid plot clutter,
        # even though multiples are technically possible.
        return next(
            collection
            for collection in ax.collections
            if isinstance(collection, collection_type)
        )


class ScalarMappableExtractorMixin:
    @staticmethod
    def extract_scalar_mappable(ax: Axes) -> ScalarMappable | None:
        """Retrieve the first collection ScalarMappable from an Axes object."""
        if ax is None or ax.get_children() is None:
            return None

        # We assume only one ScalarMappable is present to avoid plot clutter,
        # even though multiples are technically possible.
        return next(sm for sm in ax.get_children() if isinstance(sm, ScalarMappable))



from __future__ import annotations

import wrapt

from matplotlib.axes import Axes

from maidr.core.context_manager import ContextManager, BoxplotContextManager
from maidr.core.enum import PlotType
from maidr.core.figure_manager import FigureManager


@wrapt.patch_function_wrapper(Axes, "bxp")
def mpl_box(wrapped, _, args, kwargs) -> dict:
    # Don't proceed if the call is made internally by the patched function.
    if BoxplotContextManager.is_internal_context():
        plot = wrapped(*args, **kwargs)
        BoxplotContextManager.add_bxp_context(plot)
        return plot

    # Set the internal context to avoid cyclic processing.
    with ContextManager.set_internal_context():
        # Patch `ax.boxplot()` and `ax.bxp()`.
        plot = wrapped(*args, **kwargs)

    # Set the orientation of the boxplot
    if not kwargs.get("vert", True):
        orientation = "horz"
    else:
        orientation = "vert"

    # Extract the boxplot data points for MAIDR from the plot.
    ax = FigureManager.get_axes(plot)
    FigureManager.create_maidr(
        ax, PlotType.BOX, bxp_stats=plot, orientation=orientation
    )

    # Return to the caller.
    return plot


@wrapt.patch_function_wrapper("seaborn", "boxplot")
def sns_box(wrapped, _, args, kwargs) -> Axes:
    # Set the internal context to avoid cyclic processing.
    with BoxplotContextManager.set_internal_context() as bxp_context:
        # Patch `ax.boxplot()` and `ax.bxp()`.
        plot = wrapped(*args, **kwargs)
        bxp_container = bxp_context

    # Set the orientation of the boxplot
    if bxp_container.orientation() == "y" or bxp_container.orientation() == "h":
        orientation = "horz"
    else:
        orientation = "vert"

    # Extract the boxplot data points for MAIDR from the plot.
    ax = FigureManager.get_axes(bxp_container.bxp_stats())
    FigureManager.create_maidr(
        ax, PlotType.BOX, bxp_stats=bxp_container.bxp_stats(), orientation=orientation
    )

    # Return to the caller.
    return plot


def sns_infer_new_orient(wrapped, instance, args, kwargs) -> str:
    if BoxplotContextManager.is_internal_context():
        orientation = instance.orient
        BoxplotContextManager.set_bxp_orientation(orientation)

    return wrapped(*args, **kwargs)


def patch_seaborn():
    import packaging.version as version
    import seaborn

    sns_version = seaborn.__version__
    min_version = "0.12"

    if version.parse(sns_version) < version.parse(min_version):
        wrapt.wrap_function_wrapper(
            "seaborn.categorical", "_BoxPlotter.plot", sns_version
        )
    else:
        wrapt.wrap_function_wrapper(
            "seaborn.categorical",
            "_CategoricalPlotter.plot_boxes",
            sns_infer_new_orient,
        )


# Apply the appropriate patches based on the Seaborn version
patch_seaborn()



from __future__ import annotations

from matplotlib.axes import Axes
from matplotlib.container import BarContainer

from maidr.core.enum import MaidrKey, PlotType
from maidr.core.plot import MaidrPlot
from maidr.exception import ExtractionError
from maidr.util.environment import Environment
from maidr.util.mixin import (
    ContainerExtractorMixin,
    DictMergerMixin,
    LevelExtractorMixin,
)


class BarPlot(MaidrPlot, ContainerExtractorMixin, LevelExtractorMixin, DictMergerMixin):
    def __init__(self, ax: Axes) -> None:
        super().__init__(ax, PlotType.BAR)

    def _extract_axes_data(self) -> dict:
        engine = Environment.get_engine()

        base_schema = super()._extract_axes_data()
        bar_ax_schema = {}
        if engine == "js":
            bar_ax_schema = {
                MaidrKey.X: {
                    MaidrKey.LEVEL: self.extract_level(self.ax),
                },
            }
        return self.merge_dict(base_schema, bar_ax_schema)

    def _extract_plot_data(self) -> list:
        engine = Environment.get_engine()
        plot = self.extract_container(self.ax, BarContainer, include_all=True)
        data = self._extract_bar_container_data(plot)
        levels = self.extract_level(self.ax)
        if engine == "ts":
            formatted_data = []
            combined_data = (
                zip(levels, data) if plot[0].orientation == "vertical" else zip(levels, data)  # type: ignore
            )
            if len(data) == len(plot):  # type: ignore
                for x, y in combined_data:  # type: ignore
                    formatted_data.append({"x": x, "y": y})
                return formatted_data
            if len(formatted_data) == 0:
                raise ExtractionError(self.type, plot)
        if data is None:
            raise ExtractionError(self.type, plot)

        return data

    def _extract_bar_container_data(
        self, plot: list[BarContainer] | None
    ) -> list | None:
        if plot is None:
            return None

        # Since v0.13, Seaborn has transitioned from using `list[Patch]` to
        # `list[BarContainers] for plotting bar plots.
        # So, extract data correspondingly based on the level.
        # Flatten all the `list[BarContainer]` to `list[Patch]`.
        plot = [patch for container in plot for patch in container.patches]
        level = self.extract_level(self.ax)

        if len(plot) != len(level):
            return None

        # Tag the elements for highlighting.
        self._elements.extend(plot)

        return [float(patch.get_height()) for patch in plot]



__version__ = "0.15.0"

from .api import close, render, save_html, set_engine, show, stacked
from .core import Maidr
from .core.enum import PlotType
from .patch import (
    barplot,
    boxplot,
    clear,
    heatmap,
    highlight,
    histogram,
    lineplot,
    scatterplot,
)

__all__ = [
    "close",
    "render",
    "save_html",
    "show",
    "stacked",
]



from matplotlib.axes import Axes
from matplotlib.lines import Line2D

from maidr.core.enum.maidr_key import MaidrKey
from maidr.core.enum.plot_type import PlotType
from maidr.core.plot.maidr_plot import MaidrPlot
from maidr.exception.extraction_error import ExtractionError
from maidr.util.environment import Environment
from maidr.util.mixin.extractor_mixin import LineExtractorMixin


class MultiLinePlot(MaidrPlot, LineExtractorMixin):
    """
    A class for extracting and processing data from line plots.

    This class can handle both single-line and multi-line plots, extracting
    coordinate data and line identifiers. It processes matplotlib Line2D objects
    and converts them into structured data for further processing or visualization.

    Parameters
    ----------
    ax : Axes
        The matplotlib axes object containing the line plot(s).
    **kwargs : dict
        Additional keyword arguments to pass to the parent class.

    Attributes
    ----------
    type : PlotType
        Set to PlotType.LINE to identify this as a line plot.

    Notes
    -----
    - When using the JavaScript engine, only single-line plots are supported.
    - For multi-line plots, use the TypeScript engine.
    - The extracted data structure includes x, y coordinates and line identifiers
      (fill values) for each point.
    """

    def __init__(self, ax: Axes, **kwargs):
        super().__init__(ax, PlotType.LINE)

    def _get_selector(self) -> str:
        return "g[maidr='true'] > path"

    def _extract_plot_data(self) -> list[dict]:
        plot = self.extract_lines(self.ax)
        data = self._extract_line_data(plot)
        engine = Environment.get_engine()
        if engine == "js":
            if len(data) > 1:
                raise Exception(
                    "MultiLine Plot not supported in JS. Use TypeScript Engine for this plot!"
                )
            data = data[0]

        if data is None:
            raise ExtractionError(self.type, plot)

        return data

    def _extract_line_data(self, plot: list[Line2D] | None) -> list[dict] | None:
        """
        Extract data from multiple line objects.

        Parameters
        ----------
        plot : list[Line2D] | None
            List of Line2D objects to extract data from.

        Returns
        -------
        list[dict] | None
            List of dictionaries containing x,y coordinates and line identifiers,
            or None if the plot data is invalid.
        """
        if plot is None or len(plot) == 0:
            return None

        all_line_data = []

        # Process each line in the plot
        for i, line in enumerate(plot):
            if line.get_xydata() is None:
                continue

            # Tag the element for highlighting
            self._elements.append(line)

            # Extract data from this line
            line_data = [
                {
                    MaidrKey.X: float(x),
                    MaidrKey.Y: float(y),
                    MaidrKey.FILL: line.get_label(),
                }
                for x, y in line.get_xydata()  # type: ignore
            ]

            all_line_data.append(line_data)

        return all_line_data if all_line_data else None



from .extractor_mixin import (
    CollectionExtractorMixin,
    ContainerExtractorMixin,
    LevelExtractorMixin,
    LineExtractorMixin,
    ScalarMappableExtractorMixin,
)
from .merger_mixin import DictMergerMixin



from __future__ import annotations

from typing import Any, Literal

from htmltools import Tag
from matplotlib.axes import Axes
from matplotlib.container import BarContainer

from maidr.core import Maidr
from maidr.core.enum import PlotType
from maidr.core.figure_manager import FigureManager
from maidr.util.environment import Environment


def render(plot: Any) -> Tag:
    ax = FigureManager.get_axes(plot)
    maidr = FigureManager.get_maidr(ax.get_figure())
    return maidr.render()


def show(plot: Any, renderer: Literal["auto", "ipython", "browser"] = "auto") -> object:
    ax = FigureManager.get_axes(plot)
    maidr = FigureManager.get_maidr(ax.get_figure())
    return maidr.show(renderer)


def save_html(
    plot: Any, file: str, *, lib_dir: str | None = "lib", include_version: bool = True
) -> str:
    ax = FigureManager.get_axes(plot)
    maidr = FigureManager.get_maidr(ax.get_figure())
    return maidr.save_html(file, lib_dir=lib_dir, include_version=include_version)


def stacked(plot: Axes | BarContainer) -> Maidr:
    ax = FigureManager.get_axes(plot)
    return FigureManager.create_maidr(ax, PlotType.STACKED)


def close(plot: Any) -> None:
    ax = FigureManager.get_axes(plot)
    FigureManager.destroy(ax.get_figure())


def set_engine(engine: Literal["js", "ts"] = "js"):
    Environment._set_engine(engine=engine)



from __future__ import annotations

from abc import ABC, abstractmethod

from matplotlib.axes import Axes

from maidr.core.enum import MaidrKey, PlotType
from maidr.util.environment import Environment


class MaidrPlot(ABC):
    """
    Abstract base class for plots managed by the MAIDR system.

    Parameters
    ----------
    ax : Axes
        The ``matplotlib.axes.Axes`` object where the plot will be drawn.
    plot_type : PlotType
        The type of the plot to be created, as defined in the PlotType enum.

    Attributes
    ----------
    ax : Axes
        The ``matplotlib.axes.Axes`` object associated with this plot.
    type : PlotType
        The specific type of the plot.
    _schema : dict
        A dictionary containing structured data about the plot, including type, title, axes labels, and data.

    Methods
    -------
    schema()
        Returns a dictionary containing MAIDR data about the plot.
    set_id(maidr_id: str)
        Sets a unique identifier for the plot in the schema.
    """

    def __init__(self, ax: Axes, plot_type: PlotType) -> None:
        # graphic object
        self.ax = ax
        self._support_highlighting = True
        self._elements = []

        # MAIDR data
        self.type = plot_type
        self._schema = {}

    def render(self) -> dict:
        """Initialize the MAIDR schema dictionary with basic plot information."""
        maidr_schema = {
            MaidrKey.TYPE: self.type,
            MaidrKey.TITLE: self.ax.get_title(),
            MaidrKey.AXES: self._extract_axes_data(),
            MaidrKey.DATA: self._extract_plot_data(),
        }

        # Include selector only if the plot supports highlighting.
        if self._support_highlighting:
            maidr_schema[MaidrKey.SELECTOR] = self._get_selector()

        return maidr_schema

    def _get_selector(self) -> str:
        """Return the CSS selector for highlighting elements."""
        return "path[maidr='true']"

    def _extract_axes_data(self) -> dict:
        """Extract the plot's axes data"""
        engine = Environment.get_engine()
        if engine == "ts":
            return {MaidrKey.X: self.ax.get_xlabel(), MaidrKey.Y: self.ax.get_ylabel()}
        return {
            MaidrKey.X: {
                MaidrKey.LABEL: self.ax.get_xlabel(),
            },
            MaidrKey.Y: {
                MaidrKey.LABEL: self.ax.get_ylabel(),
            },
        }

    @abstractmethod
    def _extract_plot_data(self) -> list | dict:
        """Extract specific data from the plot."""
        raise NotImplementedError()

    @property
    def schema(self) -> dict:
        """Return the MAIDR schema of the plot as a dictionary."""
        if not self._schema:
            self._schema = self.render()
        return self._schema

    @property
    def elements(self) -> list:
        if not self._schema:
            self._schema = self.render()
        return self._elements

    def set_id(self, maidr_id: str) -> None:
        """Set the unique identifier for the plot within the MAIDR schema."""
        if not self._schema:
            self._schema = self.render()
        self._schema[MaidrKey.ID] = maidr_id
